#!/usr/bin/env python3

import pandas as pd
import math
import io


class CSV:
    """
    Helper class for CSV files, where named columns will be retrievable

    The class needs the data being given as list of lists. Each row (outer list) contains multiple cells (inner list).
    If the data is given JSON like – as a list of dictionaries where each row (dictionary) contains the column names as keys and the cell content as value – the second constructor-parameter has to be set `True`.
    """

    data = {}
    specs = {
        "delimiter": ";",
        "linebreak": "\n",
    }

    def __init__(self, data={}, jsonLike=False):
        if jsonLike:
            self.data = {}
            for row in data:
                try:
                    existingRowCount = len(next(iter(self.data.values())))
                except:
                    existingRowCount = 0
                keys = row.keys()
                for k in keys:
                    if k not in self.data:
                        if existingRowCount > 0:
                            self.data[k] = [None] * existingRowCount
                        else:
                            self.data[k] = []
                for k in self.data.keys():
                    if k in row:
                        self.data[k].append(row[k])
                    else:
                        self.data[k].append(None)
        else:
            self.data = data

    def setSpec(self, spec, val):
        """function to set specifications for this CSV instance"""
        self.specs[spec] = val

    def readFile(self, filepath, delimiter=None):
        """read a CSV file"""
        self.readCSV(filepath, delimiter=delimiter, file=True)

    def readCSV(self, path_or_csvstring, delimiter=None, file=False):
        """load CSV"""
        if delimiter == None:
            delimiter = self.specs["delimiter"]
        if file:
            data = pd.read_csv(path_or_csvstring, delimiter=delimiter, low_memory=False)
        else:
            data = pd.read_csv(
                io.StringIO(path_or_csvstring),
                delimiter=delimiter,
                low_memory=False,
            )
        rows = {}
        for h in data.columns:
            rows[h] = data.get(h).to_list()
        self.data = rows

    def likeJSON(self, keepEmpty=False, emptyValue=None):
        """function to turn CSV to JSON like list of dictionaries"""
        keys = self.data.keys()
        count = len(self.data[keys[0]])
        i = 0
        jsonO = []
        while i < count:
            row = {}
            for key in keys:
                value = self.data[key][i]
                if value == "":
                    if keepEmpty:
                        row[key] = emptyValue
                else:
                    row[key] = value
            jsonO.append(row)
        return jsonO

    def writeFile(self, filepath, delimiter=None, linebreak=None):
        """function to write out data of current object to a CSV file"""
        if delimiter == None:
            delimiter = self.specs["delimiter"]
        if linebreak == None:
            linebreak = self.specs["linebreak"]
        combined = [[k] + v for k, v in self.data.items()]
        cols = pd.DataFrame(combined).T.values.tolist()
        rows = []
        for r in cols:
            i = 0
            for c in r:
                if c == None:
                    r[i] = ""
                i += 1
            rows.append(
                '"{values}"'.format(
                    values='"{delim}"'.format(delim=delimiter).join(
                        [
                            str(c).replace('"', '"""')
                            if isinstance(c, str) or not math.isnan(c)
                            else ""
                            for c in r
                        ]
                    )
                )
            )
        with open(filepath, "w") as csv_file:
            csv_file.write(linebreak.join(rows))


class ColumnHelper:
    """Helper class for columns"""

    ord0 = ord("A")

    def xlsCol2Int(self, colName):
        """
        According to `A` is `0`, `Z` is `26`, `AA` is `27` and so on, this
        function is meant to translate the alphabetic “number” to an integer
        """
        val = 0
        for ch in colName:  # base-26 decoding "+1"
            val = val * 26 + ord(ch) - self.ord0 + 1
        return val - 1

    def int2xlsCol(self, colInt):
        """
        According to `A` is `0`, `Z` is `26`, `AA` is `27` and so on, this
        function is meant to translate an integer to its alphabetic “number”
        representation.
        """
        chars = []
        while True:
            if len(chars) > 0:
                colInt = colInt - 1
            ch = colInt % 26
            chars.append(chr(ch + self.ord0))
            colInt = colInt // 26
            if not colInt:
                break
        return "".join(reversed(chars))



#!/usr/bin/env python3

import requests, os, json, gc, getpass

from bs4 import BeautifulSoup

from selenium import webdriver
from selenium.common.exceptions import NoSuchElementException
from selenium.webdriver.common.desired_capabilities import DesiredCapabilities
from selenium.webdriver.common.by import By


class APIError(Exception):
    """An API Error Exception"""

    def __init__(self, status, url):
        self.status = status
        self.url = url

    def __str__(self):
        return "APIError: status={}\nURL:      {}".format(self.status, self.url)


class Browser:
    """Base browser for Selenium assisted applications"""

    browser = None
    cookies = {}
    app = {}
    lastUrl = ""
    sessionBaseHeaders = None

    appInfo = {}

    def __init__(self):
        """function to initiate the systems as needed"""
        # fetch variables from ENV
        from dotenv import load_dotenv

        load_dotenv()
        # ignore warnings, when SSL_IGNORE is set to true
        import os

        if os.getenv("SSL_IGNORE", "False").lower() in ["1", "true", "t", "y", "yes"]:
            import urllib3

            urllib3.disable_warnings()
        # gather App info and start browser
        self.gatherAppInfo()
        self.startBrowser()

    def setCookies(self, cookieDict, url=None, returnSession=False, curl=True):
        """helper function to set cookies on browser"""
        if url == None:
            url = self.app["url"]
        # parsedUrl = urllib.parse.urlparse(url)
        # domain, path = [ parsedUrl.netloc, parsedUrl.path ]
        # if path == '':
        #     path = '/'
        self.browser.get(url)
        for key, value in cookieDict.items():
            if os.getenv("PRINT_DEBUG_INFO", "False").lower() in ["1", "true", "t", "y", "yes"]:
                print("Cookie is being set:")
                print({"key": key, "value": value})
                print()
            # # does no more work since Python / Selenium Cookies can only be set for loaded URLs anymore ...
            # self.browser.add_cookie( { "name": key, "value": value, "domain": domain, "path": path } )
            self.browser.add_cookie({"name": key, "value": value})
        self.browser.refresh()
        if returnSession:
            return self.toSession(curl=curl)

    def loadCookies(self, givenCookies=None, url=None):
        """function to load cookies from JSON string ENV variable `COOKIES` – or a given Python dictionary"""
        if givenCookies == None:
            givenCookies = json.loads(os.getenv("COOKIES", "{}"))
        if len(givenCookies) > 0:
            self.setCookies(givenCookies, url)

    def gatherAppInfo(self):
        """
        function to set application information

        each `appInfo` has to be a dictionary with the following keys:
          * `env` (MANDATORY – name of the corresponding environmental variable)
          * `description` (MANDATORY – if the user is asked to enter the value, this information is shown)
          * `password` (OPTIONAL – default `False`; if set `True`, the potential user input is hidden)
          * `url` (OPTIONAL – default `False`; if set `True`, the `urlLastChar` is ensured to be last part of the value)
          * `urlLastChar` (OPTIONAL – default `/`)
          * `default` (OPTIONAL – default value for the app information; if set, the user won't be able to input the information)
        `appInfo` itself is also a dictionary and the key `url` is mandatory, the keys `user` and `passwd` are highly recommended.
        """
        self.app = {}
        for key, config in self.appInfo.items():
            value = None
            # relevant for asking for a password
            if "password" not in config:
                config["password"] = False
            # relevant for formatting urls
            if "url" not in config:
                config["url"] = False
            # relevant for formatting urls
            if "urlLastChar" not in config:
                config["urlLastChar"] = "/"
            # relevant if there is a default value for this
            # setting in the specific browser
            if "default" not in config:
                config["default"] = None
            # check for config
            if "env" in config:
                value = os.getenv(config["env"], config["default"])
            # ensure value not None
            if value == None:
                # ask for user input
                if config["password"]:
                    value = getpass.getpass("{d}: ".format(d=config["description"]))
                else:
                    value = input("{d}: ".format(d=config["description"]))
            # adjust url
            if (
                config["url"]
                and len(config["urlLastChar"]) > 0
                and value[-1] != config["urlLastChar"]
            ):
                value += config["urlLastChar"]
            # set value to app array
            self.app[key] = value

    def startBrowser(self, custom_capabilities={}):
        """Start Selenium browser"""
        btype = os.getenv("SELENIUM_BROWSER_TYPE", "FIREFOX")
        gridUrl = os.getenv("SELENIUM_GRID")

        if btype.lower() in ["chrome"]:
            capabilities = DesiredCapabilities.CHROME.copy()
        elif btype.lower() in ["opera"]:
            capabilities = DesiredCapabilities.OPERA.copy()
        elif btype.lower() in ["safari"]:
            capabilities = DesiredCapabilities.SAFARI.copy()
        else:
            capabilities = DesiredCapabilities.FIREFOX.copy()

        if os.getenv("SSL_IGNORE", "False").lower() in ["1", "true", "t", "y", "yes"]:
            capabilities["acceptInsecureCerts"] = True

        for key, value in custom_capabilities.items():
            capabilities[key] = value

        self.browser = webdriver.Remote(command_executor=gridUrl, desired_capabilities=capabilities)
        self.browser.set_script_timeout(os.getenv("SELENIUM_TIMEOUT", 30000))

    def checkElementByXpath(self, xpath):
        """helper function to check if element (XPATH) exists on page"""
        try:
            self.browser.find_element(By.XPATH, xpath)
        except NoSuchElementException:
            return False
        return True

    def checkElementByID(self, ident):
        """helper function to check if element (ID) exists on page"""
        try:
            self.browser.find_element(By.ID, ident)
        except NoSuchElementException:
            return False
        return True

    def sessionPostRequest(
        self,
        url,
        data,
        transferCookies=False,
        headers=None,
        curl=True,
        session=None,
        urlOverride=False,
    ):
        """helper function to send POST request"""
        if session == None:
            session = self.toSession(curl=curl)
        headers = self.appendSessionHeaders(headers)
        if headers is not None:
            request = session.post(url, data=data, headers=headers)
        else:
            request = session.post(url, data=data)
        if transferCookies:
            # write Cookies from POST request to Selenium Browser
            new_cookies = session.cookies.get_dict()
            if urlOverride:
                self.get(url, login=False)
                self.setCookies(new_cookies, url=url)
            else:
                self.get(request.url, login=False)
                self.setCookies(new_cookies, url=request.url)
            self.refresh()
        return request

    def sessionPutRequest(
        self, url, data, transferCookies=False, headers=None, curl=True, session=None
    ):
        """helper function to send PUT request"""
        if session == None:
            session = self.toSession(curl=curl)
        headers = self.appendSessionHeaders(headers)
        if headers is not None:
            request = session.put(url, data=data, headers=headers)
        else:
            request = session.put(url, data=data)
        if transferCookies:
            # write Cookies from POST request to Selenium Browser
            new_cookies = session.cookies.get_dict()
            self.get(request.url, login=False)
            self.setCookies(new_cookies)
            self.refresh()
        return request

    def sessionGetRequest(self, url, transferCookies=False, headers=None, curl=True, session=None):
        """helper function to send GET request"""
        if session == None:
            session = self.toSession(curl=curl)
        headers = self.appendSessionHeaders(headers)
        if headers is not None:
            request = session.get(url, headers=headers)
        else:
            request = session.get(url)
        if transferCookies:
            # write Cookies from POST request to Selenium Browser
            new_cookies = session.cookies.get_dict()
            self.get(request.url, login=False)
            self.setCookies(new_cookies)
            self.refresh()
        return request

    def apiGetInfo(self, apiUrl, returnResponse=False, session=None, curl=True):
        """Function to gather information from API URL – if `returnResponse` is True, the whole request response is returned, otherwise only the decoded answer object"""
        if session == None:
            session = self.toSession(curl=curl)
        rsp = session.get(apiUrl)
        if returnResponse:
            return rsp
        else:
            if rsp.status_code == 200:
                return json.loads(rsp.text)
            elif rsp.status_code == 404:
                return None
            else:
                raise Exception(
                    "Error while API fetch:\n======================\n\nResponse Code: {rc}\n\n{body}".format(
                        rc=rsp.status_code, body=rsp.text
                    )
                )

    def appendSessionHeaders(self, headers=None):
        """append session headers and retrieve them"""
        if headers is None:
            return self.sessionBaseHeaders
        else:
            for k, v in self.sessionBaseHeaders.items():
                if k not in headers:
                    headers[k] = v
            return headers

    def restoreCookies(self):
        """Restore cookies on Browser"""
        for cookie in self.cookies:
            self.browser.add_cookie(cookie)

    def renew(self):
        """Renew browser by closing and reopening it with restored Cookies"""
        try:
            self.browser.quit()
        except:
            pass
        self.startBrowser()
        self.browser.get(self.lastUrl)
        self.restoreCookies()

    def refresh(self):
        """Reload page in Browser"""
        currentUrl = self.browser.current_url
        self.get(currentUrl)
        # self.browser.refresh() # bricks sessionPostRequest calls ...

    def get(self, url, login=True):
        """Get url in Browser"""
        self.browser.get(url)
        if login and not self.checkLogin():
            self.login()
            self.browser.get(url)
        self.cookies = self.browser.get_cookies()
        self.lastUrl = url
        gc.collect()

    def toSession(self, curl=False):
        """Return python session"""
        sessionObj = requests.session()
        if curl:
            agent = "curl/7.54"
        else:
            agent = self.browser.execute_script("return navigator.userAgent")
        sessionObj.headers = {"User-Agent": agent}
        if len(self.cookies) > 0:
            for c in self.cookies:
                sessionObj.cookies.set(c["name"], c["value"])
        if os.getenv("SSL_IGNORE", "False").lower() in ["1", "true", "t", "y", "yes"]:
            sessionObj.verify = False
        return sessionObj

    def toSoup(self):
        """get BeautifulSoup representation of currently loaded page"""
        return BeautifulSoup(self.browser.page_source, "html.parser")

    def __del__(self):
        """magic function to close the browser instance"""
        self.browser.quit()



#!/usr/bin/env python3

import readline


class PyHistory:
    """
    Helper while development: python cli history to summarize the executed commands on CLI
    """

    showLineNumbers = False

    def __init__(self, lineNumbers=False):
        """initialize the History object"""
        self.showLineNumbers = lineNumbers

    def print(self, showLineNumbers=None):
        """Print out history"""
        if showLineNumbers == None:
            showLineNumbers = self.showLineNumbers
        if showLineNumbers:
            formatstring = "{0:4d}  {1!s}"
        else:
            formatstring = "{1!s}"
        for i in range(1, readline.get_current_history_length() + 1):
            print(formatstring.format(i, readline.get_history_item(i)))



#!/usr/bin/env python3

from setuptools import find_packages, setup
import os, pipfile
import json

pf = pipfile.load("Pipfile").data

f = open("buildconfig.json")
j = json.load(f)
build_version = j["version"]
f.close()

setup(
    name="macwinnie_pyhelpers",
    version=build_version,
    author="macwinnie",
    author_email="dev@macwinnie.me",
    license="AGPL-3.0-or-later",
    description="Helper classes for supporting Python development.",
    long_description="## Python Helpers\n\nThis toolset supports the daily coding.\n\nThe detailled documentation can be found at [GitHub](https://macwinnie.github.io/python-helpers).",
    long_description_content_type="text/markdown",
    url="https://github.com/macwinnie/python-helpers",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3 or later (AGPLv3+)",
        "Operating System :: OS Independent",
    ],
    package_dir={
        "": "src",
    },
    packages=find_packages(where="src"),
    python_requires=">={pyVersion}".format(pyVersion=pf["_meta"]["requires"]["python_version"]),
    install_requires=[
        "{package}{version}".format(package=p, version=v) if v != "*" else p
        for p, v in pf["default"].items()
    ],
    extras_requires={
        "develop": [
            "{package}{version}".format(package=p, version=v) if v != "*" else p
            for p, v in pf["develop"].items()
        ],
    },
)



#!/usr/bin/env python3

import re


class Truth:
    """
    Basic truth check class.
    """

    interpreter = None

    def __init__(self, interpretingExample=None):
        """
        Given an example object, which should be able to support interpretion
        of propopsitions, its type will be stored as interpreter. Defaults to
        `None`, could be an `Version` object out of `macwinnie_pyhelpers.Version`
        """
        self.interpreter = type(interpretingExample)

    def verify(self, proposition):
        """
        * boolean propositions will be returned as they are
        * numbers (float, integers) will return true if > 0, 0 and negatives will return false
        * `nein`, `no`, `n`, `falsch`, `false`, `f` will return `False`
        * `ja`, `j`, `yes`, `y`, `wahr`, `w`, `true`, `t` will return `True`
        """
        trueMatch = r"^((t(rue)?)|(y(es)?)|(w(ahr)?)|(j(a)?))$"
        falseMatch = r"^((f((alse)|(alsch))?)|(n((o)|(ein))?))$"
        if type(proposition) is bool:
            return proposition
        elif isinstance(proposition, (int, float)):
            return proposition > 0
        elif type(proposition) is str:
            if re.match(trueMatch, proposition.strip(), re.IGNORECASE):
                return True
            elif re.match(falseMatch, proposition.strip(), re.IGNORECASE):
                return False
            elif self.interpreter != None:
                try:
                    return self.interpret(proposition)
                except:
                    pass
        raise Exception('No interpretion found for given proposition "{}"!'.format(proposition))

    def interpret(self, proposition):
        """
        Method for deep-diving proposition interpretation.
        Will check for `P1 o P2` with `P1`, `P2` propositions and `o` an operator
        out of:
        * `<=` for “less or equal”
        * `<` for “less”
        * `>=` for “greater or equal”
        * `>` for “greater”
        * `!=` for “not equal”
        * `==` for “equal”
        """
        rx = r"((((<|>)=?)|(!|=)=))"
        cRx = re.compile(rx)
        splitting = []
        foundOperators = cRx.finditer(proposition)
        for idx, match in enumerate(foundOperators):
            s = match.start()
            e = match.end()
            splitting.append([s, e])
        checks = []
        s3 = 0
        for i, s in enumerate(splitting):
            s1 = s3
            s2 = s[0]
            s3 = s[1]
            if i != len(splitting) - 1:
                s4 = splitting[i + 1][0]
            else:
                s4 = len(proposition)
            checks.append(
                [
                    proposition[s1:s2].strip(),
                    proposition[s2:s3].strip(),
                    proposition[s3:s4].strip(),
                ]
            )
        verum = True
        for c in checks:
            o0 = self.interpreter(c[0])
            o1 = self.interpreter(c[2])
            if c[1] == "<=":
                verum = verum and o0 <= o1
            elif c[1] == "<":
                verum = verum and o0 < o1
            elif c[1] == ">=":
                verum = verum and o0 >= o1
            elif c[1] == ">":
                verum = verum and o0 > o1
            elif c[1] == "!=":
                verum = verum and o0 != o1
            elif c[1] == "==":
                verum = verum and o0 == o1
            else:
                raise Exception('"{}" is not interpretable!'.format(proposition))
        return verum


class Sentence:
    """
    String to be checked for truth.

    Binding can be done using regular brackets `(` and `)` – and
    the operators binding order.

    Operators supported are (top is more binding than bottom):
    * `¬` as `not`
    * `∧` as `and`
    * `∨` as `or` (not the letter `v`!)
    """

    sentence = None
    tsentence = None
    verifier = None
    atomized = None

    def __init__(self, sentence, interpretingExample=None, verifier=None, pretested=False):
        """
        Given a `sentence`, this classes objects will be used to verify a logical
        sentence.

        With an `interpreter` defined and `verifier` not set – may be the string
        `Version` to use the same named class from this package or `int` or
        something like that – the `verifier` from this package `Truth` will be
        instanciated.

        `verifier` has to be an instance of a class implementing a method `verify`
        which returns boolean values for string input.

        `pretested` is to be used internally - when all propositions have
        already been replaced by truthy literals.
        """
        self.sentence = sentence
        if not pretested:
            if verifier == None:
                if type(interpretingExample) == type(None):
                    self.verifier = Truth()
                else:
                    self.verifier = Truth(interpretingExample)
            else:
                self.verifier = verifier
            self.verifyAllPropositions()
        else:
            self.tsentence = sentence
        self.atomizeBraces()

    def getLiteral(self, booleanValue):
        """Somehow static function to get a literal for boolean values"""
        if booleanValue:
            return "⊤"
        else:
            return "⊥"

    def verifyAllPropositions(self):
        """Turn all propositions into `⊤` for truthy and `⊥` for false values"""
        deviders = r"([^(¬∧∨\(\)]+)"
        verified = []
        self.tsentence = self.sentence
        for p in re.findall(deviders, self.tsentence):
            x = p.strip()
            if x not in verified and len(x) > 0:
                self.tsentence = self.tsentence.replace(
                    p, self.getLiteral(self.verifier.verify(x.strip()))
                )
        self.tsentence = re.sub(r"\s", "", self.tsentence)

    def truth(self):
        """Final method that returns the truth value of a sentence"""
        while not self.isAtomic():
            # do everything to do with ¬, as it has the stronges binding
            rx = r"¬(.)"
            x = re.search(rx, self.atomized)
            if x != None:
                literal = x.group(1)
                z = not Sentence(literal, verifier=self.verifier, pretested=True).truth()
                self.atomized = self.atomized.replace(x.group(0), self.getLiteral(z))
            else:
                # do everything to do with ∧, as it's the next after ¬
                rx = r"(.)∧(.)"
                x = re.search(rx, self.atomized)
                if x != None:
                    literalA = x.group(1)
                    literalB = x.group(2)
                    boolA = Sentence(literalA, verifier=self.verifier, pretested=True).truth()
                    boolB = Sentence(literalB, verifier=self.verifier, pretested=True).truth()
                    self.atomized = self.atomized.replace(
                        x.group(0), self.getLiteral(boolA and boolB)
                    )
                else:
                    # do everything to do with ∨, as it's the next after ∧
                    rx = r"(.)∨(.)"
                    x = re.search(rx, self.atomized)
                    if x != None:
                        literalA = x.group(1)
                        literalB = x.group(2)
                        boolA = Sentence(literalA, verifier=self.verifier, pretested=True).truth()
                        boolB = Sentence(literalB, verifier=self.verifier, pretested=True).truth()
                        self.atomized = self.atomized.replace(
                            x.group(0), self.getLiteral(boolA or boolB)
                        )
        if self.atomized == "⊤":
            return True
        elif self.atomized == "⊥":
            return False
        else:
            raise Exception(
                '"{}" should be an atomic sentence – but it is not!'.format(self.sentence)
            )

    def isAtomic(self):
        """
        Check if the given sentence is a single, atomic literal so it can be translated to boolean
        """
        if len(self.atomized) == 1:
            return True
        else:
            return False

    def getBraceIdx(self, openingCharRX=r"\(", closingCharRX=r"\)"):
        """Analyze sentence for brackets and return their type and index"""
        oRx = re.compile(openingCharRX)
        cRx = re.compile(closingCharRX)

        braces = {}
        for m in oRx.finditer(self.atomized):
            c = m.group()
            if len(c) != 1:
                raise Exception("Opening only has to match single characters!")
            braces[m.start()] = "o"

        for m in cRx.finditer(self.atomized):
            c = m.group()
            if len(c) != 1:
                raise Exception("Closing only has to match single characters!")
            braces[m.start()] = "c"

        brackets = {k: braces[k] for k in sorted(braces.keys())}

        idx = list(brackets.keys())
        if len(idx) > 0:
            if brackets[idx[-1]] != "c":
                raise Exception("Last bracket has to be a closing one!")

            if brackets[idx[0]] != "o":
                raise Exception("First bracket has to be an opening one!")

        return brackets, idx

    def atomizeBraces(self):
        """Find matching outer brackets as long as existent and get literals `⊤` or `⊥` for the inner sentence"""
        if self.atomized == None:
            self.atomized = self.tsentence
        br, idx = self.getBraceIdx()
        while len(idx) > 0:
            i = 0
            z = 1
            x1 = idx[i]
            while z > 0:
                i = i + 1
                if br[idx[i]] == "o":
                    z = z + 1
                else:
                    z = z - 1
            x2 = idx[i] + 1
            toBeReplaced = self.atomized[x1:x2]
            subSentence = toBeReplaced[1:-1]
            self.atomized = self.atomized.replace(
                toBeReplaced,
                self.getLiteral(
                    Sentence(subSentence, verifier=self.verifier, pretested=True).truth()
                ),
            )
            br, idx = self.getBraceIdx()



#!/usr/bin/env python3

import re


class Version:
    """
    Helper to work with [semantic versions](https://semver.org/) – e.g. comparing them.
    """

    prefixes = []
    major = 0
    minor = 0
    patch = 0
    prefix = None
    given = None

    def __init__(self, versionString, prefix=[], removePrefix=False):
        """
        Versions could be prefixed:

        * E.g when `v1.2.3` is (or can be) used, the `prefix` parameter has to be set to `v`.
        * If one has multiple options (case sensitive!), `prefix` has to be a list of all options
        * Default is an empty list
        """
        self.given = versionString
        if type(prefix) == str:
            self.prefixes = [prefix]
        else:
            self.prefixes = prefix
        if len(self.prefixes) > 0:
            prs = sorted(self.prefixes, key=len)
            prs.reverse()
            for p in prs:
                if versionString.startswith(p):
                    if not removePrefix:
                        self.prefix = p
                    versionString = versionString[len(p) :]
                    break
        vs = versionString.split(".")
        self.major, self.minor, self.patch = map(int, vs)

    def __str__(self):
        """convert current instance to version string"""
        vString = ".".join(list(map(str, [self.major, self.minor, self.patch])))
        if self.prefix != None:
            return self.prefix + vString
        else:
            return vString

    def cmpPrepare(self, other):
        """Function to prepare comparison version"""
        if type(other) != Version:
            if type(other) == str:
                other = Version(other)
            else:
                raise Exception("Only able to compare Versions!")
        return other

    def __eq__(self, other):
        """Defines comparison method for logical operator `==`"""
        other = self.cmpPrepare(other)
        return self.major == other.major and self.minor == other.minor and self.patch == other.patch

    def __ne__(self, other):
        """Defines comparison method for logical operator `!=`"""
        return not (self == other)

    def __lt__(self, other):
        """Defines comparison method for logical operator `<`"""
        other = self.cmpPrepare(other)
        return (self.major < other.major) or (
            self.major == other.major
            and (
                self.minor < other.minor or (self.minor == other.minor and self.patch < other.patch)
            )
        )

    def __gt__(self, other):
        """Defines comparison method for logical operator `>`"""
        return not (self < other) and (self != other)

    def __le__(self, other):
        """Defines comparison method for logical operator `<=`"""
        return (self < other) or (self == other)

    def __ge__(self, other):
        """Defines comparison method for logical operator `>=`"""
        return not (self < other)

    def increase(self):
        """Increase the patch part of the current version."""
        self.patch += 1
        return str(self)

    def increaseMinor(self):
        """Increase the minor part of the current version – patch part is reset to `0`."""
        self.patch = 0
        self.minor += 1
        return str(self)

    def increaseMajor(self):
        """Increase the major part of the current version – patch and minor parts are reset to `0`."""
        self.patch = 0
        self.minor = 0
        self.major += 1
        return str(self)
